@@ -3,8 +3,7 @@
 ;; Copyright (C) 2017-2019 Johnson Denen
 ;; Copyright (C) 2021 Abhiseck Paira <abhiseckpaira@disroot.org>
 ;; Author: Johnson Denen <johnson.denen@gmail.com>
-;; Maintainer: Marty Hiatt <martianhiatus@riseup.net>
-;; Version: 1.0.0
+;; Maintainer: Marty Hiatt <mousebot@disroot.org>
 ;; Homepage: https://codeberg.org/martianh/mastodon.el
 
 ;; This file is not part of GNU Emacs.
@@ -106,7 +105,7 @@ code. Copy this code and paste it in the minibuffer prompt."
 
 (defun mastodon-auth--show-notice (notice buffer-name &optional ask)
   "Display NOTICE to user.
-NOTICE is displayed in vertical split occupying 50% of total
+By default NOTICE is displayed in vertical split occupying 50% of total
 width.  The buffer name of the buffer being displayed in the
 window is BUFFER-NAME.
 When optional argument ASK is given which should be a string, use
@@ -121,8 +120,9 @@ When ASK is absent return nil."
     (insert notice)
     (fill-region (point-min) (point-max))
     (read-only-mode)
+
     (setq window (select-window
-                  (split-window (frame-root-window) nil 'left)
+                  (split-window (frame-root-window) nil 'below)
                   t))
     (switch-to-buffer buffer t)
     (when ask
@@ -173,13 +173,13 @@ When ASK is absent return nil."
 Generate/save token if none known yet."
   (cond (mastodon-auth--token-alist
          ;; user variables are known and initialised.
-         (alist-get mastodon-instance-url mastodon-auth--token-alist))
+         (alist-get mastodon-instance-url mastodon-auth--token-alist nil nil #'string=))
         ((plist-get (mastodon-client--active-user) :access_token)
          ;; user variables need to be read from plstore.
          (push (cons mastodon-instance-url
                      (plist-get (mastodon-client--active-user) :access_token))
                mastodon-auth--token-alist)
-         (alist-get mastodon-instance-url mastodon-auth--token-alist))
+         (alist-get mastodon-instance-url mastodon-auth--token-alist nil nil #'string=))
         ((null mastodon-active-user)
          ;; user not aware of 2FA-related changes and has not set
          ;; `mastodon-active-user'. Make user aware and error out.
