@@ -1,11 +1,10 @@
 ;;; mastodon-http.el --- HTTP request/response functions for mastodon.el  -*- lexical-binding: t -*-
 
 ;; Copyright (C) 2017-2019 Johnson Denen
-;; Copyright (C) 2020-2022 Marty Hiatt
+;; Copyright (C) 2020-2024 Marty Hiatt
 ;; Author: Johnson Denen <johnson.denen@gmail.com>
-;;         Marty Hiatt <martianhiatus@riseup.net>
-;; Maintainer: Marty Hiatt <martianhiatus@riseup.net>
-;; Version: 1.0.0
+;;         Marty Hiatt <mousebot@disroot.org>
+;; Maintainer: Marty Hiatt <mousebot@disroot.org>
 ;; Homepage: https://codeberg.org/martianh/mastodon.el
 
 ;; This file is not part of GNU Emacs.
@@ -48,14 +47,15 @@
 (defconst mastodon-http--timeout 15
   "HTTP request timeout, in seconds.  Has no effect on Emacs < 26.1.")
 
-(defun mastodon-http--api (endpoint)
-  "Return Mastodon API URL for ENDPOINT."
+(defun mastodon-http--api (endpoint &optional version)
+  "Return Mastodon API URL for ENDPOINT.
+Optionally specify VERSION in format vX."
   (concat mastodon-instance-url "/api/"
-          mastodon-http--api-version "/" endpoint))
+          (or version mastodon-http--api-version) "/" endpoint))
 
-(defun mastodon-http--api-search ()
-  "Return Mastodon API url for the /search endpoint (v2)."
-  (format "%s/api/v2/search" mastodon-instance-url))
+(defun mastodon-http--api-v2 (endpoint)
+  "Return Mastodon API v2 URL for ENDPOINT."
+  (mastodon-http--api endpoint "v2"))
 
 (defun mastodon-http--response ()
   "Capture response buffer content as string."
@@ -98,10 +98,14 @@ RESPONSE if unsuccessful."
                                (mastodon-http--process-json))))
           (message "Error %s: %s" status (alist-get 'error json-response)))))))
 
-(defun mastodon-http--read-file-as-string (filename)
-  "Read a file FILENAME as a string. Used to generate image preview."
+(defun mastodon-http--read-file-as-string (filename &optional url)
+  "Read a file FILENAME as a string.
+Used to generate image preview.
+URL means FILENAME is a URL."
   (with-temp-buffer
-    (insert-file-contents filename)
+    (if url
+        (url-insert-file-contents filename)
+      (insert-file-contents filename))
     (string-to-unibyte (buffer-string))))
 
 (defmacro mastodon-http--authorized-request (method body &optional unauthenticated-p)
@@ -133,27 +137,6 @@ Used for API form data parameters that take an array."
   (cl-loop for x in array
            collect (cons param-str x)))
 
-(defun mastodon-http--post (url
-                            &optional params headers unauthenticated-p json)
-  "POST synchronously to URL, optionally with PARAMS and HEADERS.
-Authorization header is included by default unless
-UNAUTHENTICATED-P is non-nil.If JSON, encode PARAMS as JSON for
-the request data."
-  (mastodon-http--authorized-request "POST"
-    (let* ((url-request-data
-            (when params
-              (if json
-                  (json-encode params)
-                (mastodon-http--build-params-string params))))
-           (url-request-extra-headers
-            (append url-request-extra-headers ; auth set in macro
-                    (unless (assoc "Content-Type" headers) ; pleroma compat:
-                      '(("Content-Type" . "application/x-www-form-urlencoded")))
-                    headers)))
-      (with-temp-buffer
-        (mastodon-http--url-retrieve-synchronously url)))
-    unauthenticated-p))
-
 (defun mastodon-http--concat-params-to-url (url params)
   "Build a query string with PARAMS and concat to URL."
   (if params
@@ -161,6 +144,36 @@ the request data."
               (mastodon-http--build-params-string params))
     url))
 
+(defun mastodon-http--post (url
+                            &optional params headers unauthenticated-p json)
+  "POST synchronously to URL, optionally with PARAMS and HEADERS.
+Authorization header is included by default unless
+UNAUTHENTICATED-P is non-nil.
+If JSON is :json, encode PARAMS as JSON for
+the request data. If it is :raw, just use the plain params."
+  ;; NB: raw is used by `mastodon-tl--unfilter-user-languages'; not sure if
+  ;; there's a way around it?
+  (mastodon-http--authorized-request "POST"
+    (let* ((url-request-data
+            (when params
+              (cond ((eq json :json)
+                     (json-encode params))
+                    ((eq json :raw)
+                     params)
+                    (t
+                     (mastodon-http--build-params-string params)))))
+           (url-request-extra-headers
+            (append url-request-extra-headers ; auth set in macro
+                    (if json
+                        '(("Content-Type" . "application/json")
+                          ("Accept" . "application/json"))
+                      (unless (assoc "Content-Type" headers) ; pleroma compat:
+                        '(("Content-Type" . "application/x-www-form-urlencoded"))))
+                    headers)))
+      (with-temp-buffer
+        (mastodon-http--url-retrieve-synchronously url)))
+    unauthenticated-p))
+
 (defun mastodon-http--get (url &optional params silent)
   "Make synchronous GET request to URL.
 PARAMS is an alist of any extra parameters to send with the request.
@@ -176,8 +189,13 @@ PARAMS is an alist of any extra parameters to send with the request.
 SILENT means don't message.
 NO-HEADERS means don't collect http response headers.
 VECTOR means return json arrays as vectors."
-  (with-current-buffer (mastodon-http--get url params silent)
-    (mastodon-http--process-response no-headers vector)))
+(let ((buf (mastodon-http--get url params silent)))
+  ;; --get can return nil if instance unresponsive:
+  (if (not buf)
+      (user-error "Looks like the server response borked. \
+Is your instance up?")
+    (with-current-buffer buf
+      (mastodon-http--process-response no-headers vector)))))
 
 (defun mastodon-http--get-json (url &optional params silent vector)
   "Return only JSON data from URL request.
@@ -225,7 +243,9 @@ Callback to `mastodon-http--get-response-async', usually
             ;; ideally we should handle the status code in here rather than
             ;; this crappy hack?
             ((string-prefix-p "\n<" json-string) ; html hack
-             (mastodon-http--render-html-err json-string))
+             ;; NB: in this case, process-response returns t!:
+             (mastodon-http--render-html-err json-string)
+             nil) ;; return nil instead of t
             ;; if no json or html, maybe we have a plain string error message
             ;; (misskey does this, but there are probably better ways to do
             ;; this):
@@ -233,11 +253,14 @@ Callback to `mastodon-http--get-response-async', usually
                       (string-prefix-p "\n[" json-string)))
              (error "%s" json-string))
             (t
+             ;; instance may return error in JSON e.g. ((error . "Record not
+             ;; found")) for a null endpoint. but we don't error here because
+             ;; sometimes we just want to check for such an error in an
+             ;; if/cond.
              `(,(json-read-from-string json-string) . ,headers))))))
 
 (defun mastodon-http--process-headers ()
   "Return an alist of http response headers."
-  (switch-to-buffer (current-buffer))
   (goto-char (point-min))
   (let* ((head-str (buffer-substring-no-properties
                     (point-min)
@@ -279,11 +302,26 @@ Optionally specify the PARAMS to send."
   (with-current-buffer (mastodon-http--patch url params)
     (mastodon-http--process-json)))
 
-(defun mastodon-http--patch (base-url &optional params)
-  "Make synchronous PATCH request to BASE-URL.
-Optionally specify the PARAMS to send."
+(defun mastodon-http--patch (url &optional params json)
+  "Make synchronous PATCH request to URL.
+Optionally specify the PARAMS to send.
+JSON means send params as JSON data."
   (mastodon-http--authorized-request "PATCH"
-    (let ((url (mastodon-http--concat-params-to-url base-url params)))
+    ;; NB: unlike POST, PATCHing only works if we use query params!
+    ;; so here, unless JSON arg, we use query params and do not set
+    ;; `url-request-data'. this is probably an error, i don't understand it.
+    (let* ((url-request-data
+            (when (and params json)
+              (encode-coding-string
+               (json-encode params) 'utf-8)))
+           ;; (mastodon-http--build-params-string params))))
+           (url (unless json
+                  (mastodon-http--concat-params-to-url url params)))
+           (headers (when json
+                      '(("Content-Type" . "application/json")
+                        ("Accept" . "application/json"))))
+           (url-request-extra-headers
+            (append url-request-extra-headers headers)))
       (mastodon-http--url-retrieve-synchronously url))))
 
  ;; Asynchronous functions
@@ -335,18 +373,27 @@ The upload is asynchronous. On succeeding,
 item uploaded, and `mastodon-toot--update-status-fields' is run."
   (let* ((file (file-name-nondirectory filename))
          (request-backend 'curl)
+         (desc `(("description" . ,caption)))
          (cb (cl-function
               (lambda (&key data &allow-other-keys)
                 (when data
-                  (push (alist-get 'id data)
-                        mastodon-toot--media-attachment-ids) ; add ID to list
-                  (message (alist-get 'id data))
-                  (message "Uploading %s... (done)" file)
-                  (mastodon-toot--update-status-fields))))))
+                  (let* ((id (alist-get 'id data)))
+                    ;; update ids:
+                    (push id mastodon-toot--media-attachment-ids)
+                    ;; pleroma, PUT the description:
+                    ;; this is how the mangane akkoma web client does it
+                    ;; and it seems easier than the other options!
+                    (when (and caption
+                               (not (string= caption (alist-get 'description data))))
+                      (let ((url (mastodon-http--api (format "media/%s" id))))
+                        ;; (message "PUTting image description")
+                        (mastodon-http--put url desc)))
+                    (message "Uploading %s... (done)" file)
+                    (mastodon-toot--update-status-fields)))))))
     (request
       url
       :type "POST"
-      :params `(("description" . ,caption))
+      :params desc
       :files `(("file" . (,file :file ,filename
                                 :mime-type "multipart/form-data")))
       :parser 'json-read
