@@ -1,9 +1,8 @@
 ;;; mastodon-search.el --- Search functions for mastodon.el  -*- lexical-binding: t -*-
 
 ;; Copyright (C) 2017-2019 Marty Hiatt
-;; Author: Marty Hiatt <martianhiatus@riseup.net>
-;; Maintainer: Marty Hiatt <martianhiatus@riseup.net>
-;; Version: 1.0.0
+;; Author: Marty Hiatt <mousebot@disroot.org>
+;; Maintainer: Marty Hiatt <mousebot@disroot.org>
 ;; Homepage: https://codeberg.org/martianh/mastodon.el
 
 ;; This file is not part of GNU Emacs.
@@ -29,8 +28,8 @@
 
 ;;; Code:
 (require 'json)
-(eval-when-compile
-  (require 'mastodon-tl))
+(require 'mastodon-tl)
+(require 'mastodon-widget)
 
 (autoload 'mastodon-auth--access-token "mastodon-auth")
 (autoload 'mastodon-http--api "mastodon-http")
@@ -45,7 +44,7 @@
 (autoload 'mastodon-tl--timeline "mastodon-tl")
 (autoload 'mastodon-tl--toot "mastodon-tl")
 (autoload 'mastodon-tl--buffer-property "mastodon-tl")
-(autoload 'mastodon-http--api-search "mastodon-http")
+(autoload 'mastodon-http--api-v2 "mastodon-http")
 
 (defvar mastodon-toot--completion-style-for-mentions)
 (defvar mastodon-instance-url)
@@ -65,11 +64,13 @@
 Returns a nested list containing user handle, display name, and URL."
   (let* ((url (mastodon-http--api "accounts/search"))
          (response
-          (if (equal mastodon-toot--completion-style-for-mentions "following")
-              (mastodon-http--get-json
-               url `(("q" . ,query) ("following" . "true"))
-               :silent)
-            (mastodon-http--get-json url `(("q" . ,query)) :silent))))
+          (mastodon-http--get-json
+           url
+           `(("q" . ,query) ;; NB: nil can break params (but works for me)
+             ,(when (string= "following"
+                           mastodon-toot--completion-style-for-mentions)
+                '("following" . "true")))
+           :silent)))
     (mapcar #'mastodon-search--get-user-info-@ response)))
 
 ;; functions for tags completion:
@@ -77,7 +78,7 @@ Returns a nested list containing user handle, display name, and URL."
 (defun mastodon-search--search-tags-query (query)
   "Return an alist containing tag strings plus their URLs.
 QUERY is the string to search."
-  (let* ((url (mastodon-http--api-search))
+  (let* ((url (mastodon-http--api-v2 "search"))
          (params `(("q" . ,query) ("type" . "hashtags")))
          (response (mastodon-http--get-json url params :silent))
          (tags (alist-get 'hashtags response)))
@@ -97,48 +98,88 @@ QUERY is the string to search."
   (mastodon-search--view-trending "statuses"
                                   #'mastodon-tl--timeline))
 
+(defun mastodon-search--trending-links ()
+  "Display a list of links trending on your instance."
+  (interactive)
+  (mastodon-search--view-trending "links"
+                                  #'mastodon-search--render-links))
+
+(defun mastodon-search--render-links (links)
+  "Render trending LINKS."
+  (cl-loop for l in links
+           do (mastodon-search--render-link l)))
+
+(defun mastodon-search--render-link (link)
+  "Render a trending LINK."
+  (let-alist link
+    (insert
+     (propertize
+      (mastodon-tl--render-text
+       (concat "<a href=\"" .url "\">" .url "</a>\n" .title)
+       link)
+      'item-type 'link
+      'item-json link
+      'shr-url .url
+      'byline t ;; nav
+      'help-echo
+      (substitute-command-keys
+       "\\[`mastodon-search--load-link-posts'] to view a link's timeline"))
+     ;; TODO: display card link author here
+     "\n\n")))
+
+(defun mastodon-search--load-link-posts ()
+  "Load timeline of posts containing link at point."
+  (interactive)
+  (let* ((url (mastodon-tl--property 'shr-url)))
+    (mastodon-tl--link-timeline url)))
+
 (defun mastodon-search--view-trending (type print-fun)
   "Display a list of tags trending on your instance.
 TYPE is a string, either tags, statuses, or links.
 PRINT-FUN is the function used to print the data from the response."
-  (let* ((url (mastodon-http--api
-               (format "trends/%s" type)))
+  (let* ((url (mastodon-http--api (format "trends/%s" type)))
          ;; max for statuses = 40, for others = 20
-         (limit (if (equal type "statuses")
+         (limit (if (string= type "statuses")
                     '("limit" . "40")
                   '("limit" . "20")))
          (offset '(("offset" . "0")))
          (params (push limit offset))
          (data (mastodon-http--get-json url params))
-         (buffer (get-buffer-create (format "*mastodon-trending-%s*" type))))
+         (buffer (get-buffer-create
+                  (format "*mastodon-trending-%s*" type))))
     (with-mastodon-buffer buffer #'mastodon-mode nil
       (mastodon-tl--set-buffer-spec (buffer-name buffer)
                                     (format "trends/%s" type)
-                                    print-fun nil
-                                    params)
+                                    print-fun nil params)
       (mastodon-search--insert-heading "trending" type)
       (funcall print-fun data)
-      (unless (equal type "statuses")
+      (unless (string= type "statuses")
         (goto-char (point-min))))))
 
 ;; functions for mastodon search
 
-(defun mastodon-search--insert-heading (heading &optional type)
-  "Format HEADING as a heading.
+(defun mastodon-search--insert-heading (str &optional type)
+  "Insert STR as a heading.
 Optionally add string TYPE after HEADING."
   (insert
-   (mastodon-tl--set-face (concat "\n " mastodon-tl--horiz-bar "\n "
-                                  (upcase heading) " "
-                                  (if type (upcase type) "") "\n"
-                                  " " mastodon-tl--horiz-bar "\n")
-                          'success)))
+   (mastodon-search--format-heading str type)))
+
+(defun mastodon-search--format-heading (str &optional type no-newline)
+  "Format STR as a heading.
+Optionally add string TYPE after HEADING.
+NO-NEWLINE means don't add add a newline at end."
+  (mastodon-tl--set-face
+   (concat "\n " mastodon-tl--horiz-bar "\n "
+           (upcase str) " "
+           (when type (upcase type)) "\n"
+           " " mastodon-tl--horiz-bar (unless no-newline "\n"))
+   'success))
 
 (defvar mastodon-search-types
   '("statuses" "accounts" "hashtags"))
 
-(defun mastodon-search--query (query
-                               &optional type limit
-                               following account-id offset)
+(defun mastodon-search--query (query &optional type limit
+                                     following account-id offset)
   "Prompt for a search QUERY and return accounts, statuses, and hashtags.
 TYPE is a member of `mastodon-search-types'.
 LIMIT is a number as string, up to 40, with 40 the default.
@@ -149,16 +190,14 @@ OFFSET is a number as string, means to skip that many results. It
 is used for pagination."
   ;; TODO: handle no results
   (interactive "sSearch mastodon for: ")
-  (let* ((url (mastodon-http--api-search))
-         (following (when (or following
-                              (equal current-prefix-arg '(4)))
+  (let* ((url (mastodon-http--api-v2 "search"))
+         (following (when (or following (equal current-prefix-arg '(4)))
                       "true"))
          (type (or type
                    (if (equal current-prefix-arg '(4))
                        "accounts" ; if FOLLOWING, must be "accounts"
                      (completing-read "Search type: "
-                                      mastodon-search-types
-                                      nil t))))
+                                      mastodon-search-types nil :match))))
          (limit (or limit "40"))
          (offset (or offset "0"))
          (buffer (format "*mastodon-search-%s-%s*" type query))
@@ -170,26 +209,28 @@ is used for pagination."
                               ,(when following `("following" . ,following))
                               ,(when account-id `("account_id" . ,account-id)))))
          (response (mastodon-http--get-json url params))
-         (accts (when (equal type "accounts")
-                  (alist-get 'accounts response)))
-         (tags (when (equal type "hashtags")
-                 (alist-get 'hashtags response)))
-         (statuses (when (equal type "statuses")
-                     (alist-get 'statuses response))))
+         (items (alist-get (intern type) response)))
     (with-mastodon-buffer buffer #'mastodon-mode nil
       (mastodon-search-mode)
-      (mastodon-search--insert-heading type)
-      ;; user results:
-      (cond ((equal type "accounts")
-             (mastodon-search--render-response accts type buffer params
+      (mastodon-search--insert-heading "search")
+      (mastodon-widget--create
+       "Results"
+       '(accounts hashtags statuses)
+       (intern type)
+       (lambda (widget &rest _ignore)
+         (let ((value (widget-value widget)))
+           (mastodon-search--query query (symbol-name value)))))
+      (insert "\n\n")
+      (cond ((string= type "accounts")
+             (mastodon-search--render-response items type buffer params
                                                'mastodon-views--insert-users-propertized-note
                                                'mastodon-views--insert-users-propertized-note))
-            ((equal type "hashtags")
-             (mastodon-search--render-response tags type buffer params
+            ((string= type "hashtags")
+             (mastodon-search--render-response items type buffer params
                                                'mastodon-search--print-tags
                                                'mastodon-search--print-tags))
-            ((equal type "statuses")
-             (mastodon-search--render-response statuses type buffer params
+            ((string= type "statuses")
+             (mastodon-search--render-response items type buffer params
                                                #'mastodon-tl--timeline
                                                #'mastodon-tl--timeline)))
       (goto-char (point-min))
@@ -199,7 +240,7 @@ is used for pagination."
 
 (defun mastodon-search-insert-no-results (&optional thing)
   "Insert a no results message for object THING."
-  (let ((thing (or thing "nothing")))
+  (let ((thing (or thing "items")))
     (insert
      (propertize (format "Looks like search returned no %s." thing)
                  'face 'font-lock-comment-face))))
@@ -211,31 +252,29 @@ BUFFER, PARAMS, and UPDATE-FUN are for `mastodon-tl--buffer-spec'."
   (if (not data)
       (mastodon-search-insert-no-results type)
     (funcall insert-fun data))
-  ;; (mapc #'mastodon-tl--toot data))
   (mastodon-tl--set-buffer-spec buffer "search"
-                                update-fun
-                                nil params))
+                                update-fun nil params))
 
 (defun mastodon-search--buf-type ()
   "Return search buffer type, a member of `mastodon-search-types'."
   ;; called in `mastodon-tl--get-buffer-type'
   (let* ((spec (mastodon-tl--buffer-property 'update-params)))
-    (alist-get "type" spec nil nil #'equal)))
+    (alist-get "type" spec nil nil #'string=)))
 
 (defun mastodon-search--query-cycle ()
   "Cycle through search types: accounts, hashtags, and statuses."
   (interactive)
   (let* ((spec (mastodon-tl--buffer-property 'update-params))
-         (type (alist-get "type" spec nil nil #'equal))
-         (query (alist-get "q" spec nil nil #'equal)))
-    (cond ((equal type "hashtags")
+         (type (alist-get "type" spec nil nil #'string=))
+         (query (alist-get "q" spec nil nil #'string=)))
+    (cond ((string= type "hashtags")
            (mastodon-search--query query "accounts"))
-          ((equal type "accounts")
+          ((string= type "accounts")
            (mastodon-search--query query "statuses"))
-          ((equal type "statuses")
+          ((string= type "statuses")
            (mastodon-search--query query "hashtags")))))
 
-(defun mastodon-serach--query-accounts-followed (query)
+(defun mastodon-search--query-accounts-followed (query)
   "Run an accounts search QUERY, limited to your followers."
   (interactive "sSearch mastodon for: ")
   (mastodon-search--query query "accounts" :following))
@@ -272,9 +311,8 @@ If NOTE is non-nil, include user's profile note. This is also
                   'mastodon-handle (concat "@" (cadr user))
 		  'help-echo (concat "Browse user profile of @" (cadr user)))
       " : \n"
-      (if note
-          (mastodon-tl--render-text (cadddr user) acct)
-        "")
+      (when note
+        (mastodon-tl--render-text (cadddr user) acct))
       "\n")
      'item-json acct))) ; for compat w other processing functions
 
@@ -302,9 +340,7 @@ If NOTE is non-nil, include user's profile note. This is also
 
 (defun mastodon-search--get-user-info (account)
   "Get user handle, display name, account URL and profile note from ACCOUNT."
-  (list (if (not (string-empty-p (alist-get 'display_name account)))
-            (alist-get 'display_name account)
-          (alist-get 'username account))
+  (list (mastodon-tl--display-or-uname account)
         (alist-get 'acct account)
         (alist-get 'url account)
         (alist-get 'note account)))
@@ -314,26 +350,29 @@ If NOTE is non-nil, include user's profile note. This is also
   (list (alist-get 'name tag)
         (alist-get 'url tag)))
 
-(defun mastodon-search--get-status-info (status)
-  "Get ID, timestamp, content, and spoiler from STATUS."
-  (list (alist-get 'id status)
-        (alist-get 'created_at status)
-        (alist-get 'spoiler_text status)
-        (alist-get 'content status)))
+;; These functions are all unused!
 
-(defun mastodon-search--id-from-status (status)
-  "Fetch the id from a STATUS returned by a search call to the server.
-We use this to fetch the complete status from the server."
-  (alist-get 'id status))
+;; (defun mastodon-search--get-status-info (status)
+;;   "Get ID, timestamp, content, and spoiler from STATUS."
+;;   (list (alist-get 'id status)
+;;         (alist-get 'created_at status)
+;;         (alist-get 'spoiler_text status)
+;;         (alist-get 'content status)))
 
-(defun mastodon-search--full-status-from-id (id)
-  "Fetch the full status with id ID from the server.
-This allows us to access the full account etc. details and to
-render them properly."
-  (let* ((url (concat mastodon-instance-url "/api/v1/statuses/" (mastodon-tl--as-string id)))
-         (json (mastodon-http--get-json url)))
-    json))
+;; (defun mastodon-search--id-from-status (status)
+;;   "Fetch the id from a STATUS returned by a search call to the server.
+;; We use this to fetch the complete status from the server."
+;;   (alist-get 'id status))
 
+;; (defun mastodon-search--full-status-from-id (id)
+;;   "Fetch the full status with id ID from the server.
+;; This allows us to access the full account etc. details and to
+;; render them properly."
+;;   (let* ((url (mastodon-http--api (format "statuses/%s" id)))
+;;          ;; (concat mastodon-instance-url "/api/v1/statuses/"
+;;          ;; (mastodon-tl--as-string id)))
+;;          (json (mastodon-http--get-json url)))
+;;     json))
 
 (defvar mastodon-search-mode-map
   (let ((map (make-sparse-keymap)))
@@ -350,6 +389,5 @@ This minor mode is used for mastodon search pages to adds a keybinding."
   :group 'mastodon
   :global nil)
 
-
 (provide 'mastodon-search)
 ;;; mastodon-search.el ends here
