@@ -4,10 +4,10 @@
 ;; Copyright (C) 2020-2022 Marty Hiatt
 ;; Copyright (C) 2021 Abhiseck Paira <abhiseckpaira@disroot.org>
 ;; Author: Johnson Denen <johnson.denen@gmail.com>
-;;         Marty Hiatt <martianhiatus@riseup.net>
-;; Maintainer: Marty Hiatt <martianhiatus@riseup.net>
-;; Version: 1.0.12
-;; Package-Requires: ((emacs "27.1") (request "0.3.0") (persist "0.4"))
+;;         Marty Hiatt <mousebot@disroot.org>
+;; Maintainer: Marty Hiatt <mousebot@disroot.org>
+;; Version: 1.1.8
+;; Package-Requires: ((emacs "28.1") (request "0.3.0") (persist "0.4") (tp "0.6"))
 ;; Homepage: https://codeberg.org/martianh/mastodon.el
 
 ;; This file is not part of GNU Emacs.
@@ -32,19 +32,21 @@
 ;; mastodon.el is a client for fediverse services that implement the Mastodon
 ;; API. See <https://github.com/mastodon/mastodon>.
 
-;; See the readme file at https://codeberg.org/martianh/mastodon.el for set up
-;; and usage details.
+;; For set up and usage details, see the Info documentation, or the readme
+;; file at <https://codeberg.org/martianh/mastodon.el>.
 
 ;;; Code:
-(require 'cl-lib) ; for `cl-some' call in mastodon
+(require 'cl-lib) ; for `cl-some' call in `mastodon'
 (eval-when-compile (require 'subr-x))
-(require 'mastodon-http)
-(require 'mastodon-toot)
-(require 'mastodon-search)
 (require 'url)
 (require 'thingatpt)
 (require 'shr)
 
+(require 'mastodon-http)
+(require 'mastodon-toot)
+(require 'mastodon-search)
+(require 'mastodon-transient)
+
 (declare-function discover-add-context-menu "discover")
 (declare-function emojify-mode "emojify")
 (declare-function request "request")
@@ -97,7 +99,10 @@
 (autoload 'mastodon-views--view-instance-description "mastodon-views")
 (autoload 'mastodon-views--view-lists "mastodon-views")
 (autoload 'mastodon-views--view-scheduled-toots "mastodon-views")
+(autoload 'mastodon-tl--dm-user "mastodon-tl")
+(autoload 'mastodon-tl--scroll-up-command "mastodon-tl")
 (autoload 'special-mode "simple")
+(autoload 'mastodon-tl--thread-do "mastodon-tl")
 
 (defvar mastodon-tl--highlight-current-toot)
 (defvar mastodon-notifications--map)
@@ -108,8 +113,8 @@
   :group 'external)
 
 (defcustom mastodon-instance-url "https://mastodon.social"
-  "Base URL for the Mastodon instance you want to be active.
-For example, if your mastodon username is
+  "Base URL for the fediverse instance you want to be active.
+For example, if your username is
 \"example_user@social.instance.org\", and you want this account
 to be active, the value of this variable should be
 \"https://social.instance.org\".
@@ -123,7 +128,7 @@ changes to take effect."
 
 (defcustom mastodon-active-user nil
   "Username of the active user.
-For example, if your mastodon username is
+For example, if your username is
 \"example_user@social.instance.org\", and you want this account
 to be active, the value of this variable should be
 \"example_user\".
@@ -142,82 +147,132 @@ The default value \"%F %T\" prints ISO8601-style YYYY-mm-dd HH:MM:SS.
 Use. e.g. \"%c\" for your locale's date and time format."
   :type 'string)
 
+(defcustom mastodon-use-emojify nil
+  "Whether to use emojify.el to display emojis.
+From version 28, Emacs can display emojis natively. But
+currently, it doesn't seem to have a way to handle custom emoji,
+while emojify,el has this feature and mastodon.el implements it."
+  :type 'boolean)
+
+;; notifications customizes
+;; moved here because we can load notifs without first loading mastodon.el
+;; or mastodon-notifications.el
+
+(defcustom mastodon-profile-note-in-foll-reqs t
+  "If non-nil, show a user's profile note in follow request notifications."
+  :type '(boolean))
+
+(defcustom mastodon-profile-note-in-foll-reqs-max-length nil
+  "The max character length for user profile note in follow requests.
+Profile notes are only displayed if
+`mastodon-profile-note-in-foll-reqs' is non-nil.
+If unset, profile notes of any size will be displayed, which may
+make them unweildy."
+  :type '(integer))
+
+(defcustom mastodon-images-in-notifs nil
+  "Whether to display attached images in notifications."
+  :type '(boolean))
+
+(defcustom mastodon-group-notifications nil
+  "Whether to use grouped notifications.
+Requires that your instance actually implements grouped notifications.
+Mastodon implemented them in 4.3."
+  :type '(boolean))
+
+(defcustom mastodon-notifications-grouped-names-count 2
+  "The number of notification authors to display.
+A count of 2 for example means to display like so: \"Bob, Jenny
+and X others...\"."
+  :type '(integer))
+
+(defun mastodon-kill-window ()
+  "Quit window and delete helper."
+  (interactive)
+  (quit-window 'kill))
+
 (defvar mastodon-mode-map
   (let ((map (make-sparse-keymap)))
     ;; navigation inside a timeline
-    (define-key map (kbd "n") #'mastodon-tl--goto-next-item)
-    (define-key map (kbd "p") #'mastodon-tl--goto-prev-item)
-    (define-key map (kbd "M-n") #'mastodon-tl--next-tab-item)
-    (define-key map (kbd "M-p") #'mastodon-tl--previous-tab-item)
-    (define-key map [?\t] #'mastodon-tl--next-tab-item)
-    (define-key map [backtab] #'mastodon-tl--previous-tab-item)
-    (define-key map [?\S-\t] #'mastodon-tl--previous-tab-item)
-    (define-key map [?\M-\t] #'mastodon-tl--previous-tab-item)
-    (define-key map (kbd "l") #'recenter-top-bottom)
+    (define-key map (kbd "n")      #'mastodon-tl--goto-next-item)
+    (define-key map (kbd "p")      #'mastodon-tl--goto-prev-item)
+    (define-key map (kbd "M-n")    #'mastodon-tl--next-tab-item)
+    (define-key map (kbd "M-p")    #'mastodon-tl--previous-tab-item)
+    (define-key map [?\t]          #'mastodon-tl--next-tab-item)
+    (define-key map [backtab]      #'mastodon-tl--previous-tab-item)
+    (define-key map [?\S-\t]       #'mastodon-tl--previous-tab-item)
+    (define-key map [?\M-\t]       #'mastodon-tl--previous-tab-item)
+    (define-key map (kbd "l")      #'recenter-top-bottom)
     ;; navigation between timelines
-    (define-key map (kbd "#") #'mastodon-tl--get-tag-timeline)
-    (define-key map (kbd "\"") #'mastodon-tl--list-followed-tags)
-    (define-key map (kbd "'") #'mastodon-tl--followed-tags-timeline)
-    (define-key map (kbd "A") #'mastodon-profile--get-toot-author)
-    (define-key map (kbd "F") #'mastodon-tl--get-federated-timeline)
-    (define-key map (kbd "H") #'mastodon-tl--get-home-timeline)
-    (define-key map (kbd "L") #'mastodon-tl--get-local-timeline)
-    (define-key map (kbd "N") #'mastodon-notifications-get)
-    (define-key map (kbd "@") #'mastodon-notifications--get-mentions)
-    (define-key map (kbd "P") #'mastodon-profile--show-user)
-    (define-key map (kbd "s") #'mastodon-search--query)
-    (define-key map (kbd "/") #'mastodon-switch-to-buffer)
+    (define-key map (kbd "#")      #'mastodon-tl--get-tag-timeline)
+    (define-key map (kbd "\"")     #'mastodon-tl--list-followed-tags)
+    (define-key map (kbd "'")      #'mastodon-tl--followed-tags-timeline)
+    (define-key map (kbd "A")      #'mastodon-profile--get-toot-author)
+    (define-key map (kbd "F")      #'mastodon-tl--get-federated-timeline)
+    (define-key map (kbd "H")      #'mastodon-tl--get-home-timeline)
+    (define-key map (kbd "L")      #'mastodon-tl--get-local-timeline)
+    (define-key map (kbd "N")      #'mastodon-notifications-get)
+    (define-key map (kbd "@")      #'mastodon-notifications--get-mentions)
+    (define-key map (kbd "P")      #'mastodon-profile--show-user)
+    (define-key map (kbd "s")      #'mastodon-search--query)
+    (define-key map (kbd "/")      #'mastodon-switch-to-buffer)
+    (define-key map (kbd "\\")     #'mastodon-tl--get-remote-local-timeline)
     ;; quitting mastodon
-    (define-key map (kbd "q") #'kill-current-buffer)
-    (define-key map (kbd "Q") #'kill-buffer-and-window)
-    (define-key map (kbd "M-C-q") #'mastodon-kill-all-buffers)
+    (define-key map (kbd "q")      #'kill-current-buffer)
+    (define-key map (kbd "Q")      #'mastodon-kill-window)
+    (define-key map (kbd "M-C-q")  #'mastodon-kill-all-buffers)
     ;; toot actions
-    (define-key map (kbd "c") #'mastodon-tl--toggle-spoiler-text-in-toot)
-    (define-key map (kbd "b") #'mastodon-toot--toggle-boost)
-    (define-key map (kbd "f") #'mastodon-toot--toggle-favourite)
-    (define-key map (kbd "k") #'mastodon-toot--toggle-bookmark)
-    (define-key map (kbd "r") #'mastodon-toot--reply)
-    (define-key map (kbd "C") #'mastodon-toot--copy-toot-url)
-    (define-key map (kbd "v") #'mastodon-tl--poll-vote)
-    (define-key map (kbd "E") #'mastodon-toot--view-toot-edits)
-    (define-key map (kbd "T") #'mastodon-tl--thread)
-    (define-key map (kbd "m") #'mastodon-tl--dm-user)
+    (define-key map (kbd "c")      #'mastodon-tl--toggle-spoiler-text-in-toot)
+    (define-key map (kbd "b")      #'mastodon-toot--toggle-boost)
+    (define-key map (kbd "f")      #'mastodon-toot--toggle-favourite)
+    (define-key map (kbd "k")      #'mastodon-toot--toggle-bookmark)
+    (define-key map (kbd "r")      #'mastodon-toot--reply)
+    (define-key map (kbd "C")      #'mastodon-toot--copy-toot-url)
+    (define-key map (kbd "o")      #'mastodon-toot--browse-toot-url)
+    (define-key map (kbd "v")      #'mastodon-tl--poll-vote)
+    (define-key map (kbd "E")      #'mastodon-toot--view-toot-edits)
+    (define-key map (kbd "T")      #'mastodon-tl--thread)
+    (define-key map (kbd "RET")    #'mastodon-tl--thread)
+    (define-key map (kbd "m")      #'mastodon-tl--dm-user)
     (when (require 'lingva nil :no-error)
-      (define-key map (kbd "a") #'mastodon-toot--translate-toot-text))
-    (define-key map (kbd ",") #'mastodon-toot--list-toot-favouriters)
-    (define-key map (kbd ".") #'mastodon-toot--list-toot-boosters)
-    (define-key map (kbd ";") #'mastodon-views--view-instance-description)
+      (define-key map (kbd "a")    #'mastodon-toot--translate-toot-text))
+    (define-key map (kbd ",")      #'mastodon-toot--list-favouriters)
+    (define-key map (kbd ".")      #'mastodon-toot--list-boosters)
+    (define-key map (kbd ";")      #'mastodon-views--view-instance-description)
     ;; override special mode binding
-    (define-key map (kbd "g") #'undefined)
-    (define-key map (kbd "g") #'mastodon-tl--update)
+    (define-key map (kbd "g")      #'undefined)
+    (define-key map (kbd "g")      #'mastodon-tl--update)
     ;; this is now duplicated by 'g', cd remove/use for else:
-    (define-key map (kbd "u") #'mastodon-tl--update)
+    (define-key map (kbd "u")      #'mastodon-tl--update)
     ;; own toot actions:
-    (define-key map (kbd "t") #'mastodon-toot)
-    (define-key map (kbd "d") #'mastodon-toot--delete-toot)
-    (define-key map (kbd "D") #'mastodon-toot--delete-and-redraft-toot)
-    (define-key map (kbd "i") #'mastodon-toot--pin-toot-toggle)
-    (define-key map (kbd "e") #'mastodon-toot--edit-toot-at-point)
+    (define-key map (kbd "t")      #'mastodon-toot)
+    (define-key map (kbd "d")      #'mastodon-toot--delete-toot)
+    (define-key map (kbd "D")      #'mastodon-toot--delete-and-redraft-toot)
+    (define-key map (kbd "i")      #'mastodon-toot--pin-toot-toggle)
+    (define-key map (kbd "e")      #'mastodon-toot--edit-toot-at-point)
     ;; user actions
-    (define-key map (kbd "W") #'mastodon-tl--follow-user)
-    (define-key map (kbd "C-S-W") #'mastodon-tl--unfollow-user)
-    (define-key map (kbd "B") #'mastodon-tl--block-user)
-    (define-key map (kbd "C-S-B") #'mastodon-tl--unblock-user)
-    (define-key map (kbd "M") #'mastodon-tl--mute-user)
-    (define-key map (kbd "C-S-M") #'mastodon-tl--unmute-user)
-    (define-key map (kbd "Z") #'mastodon-tl--report-to-mods)
+    (define-key map (kbd "W")      #'mastodon-tl--follow-user)
+    (define-key map (kbd "C-S-W")  #'mastodon-tl--unfollow-user)
+    (define-key map (kbd "B")      #'mastodon-tl--block-user)
+    (define-key map (kbd "C-S-B")  #'mastodon-tl--unblock-user)
+    (define-key map (kbd "M")      #'mastodon-tl--mute-user)
+    (define-key map (kbd "C-S-M")  #'mastodon-tl--unmute-user)
+    (define-key map (kbd "Z")      #'mastodon-tl--report-to-mods)
     ;; own profile
-    (define-key map (kbd "O") #'mastodon-profile--my-profile)
-    (define-key map (kbd "U") #'mastodon-profile--update-user-profile-note)
-    (define-key map (kbd "V") #'mastodon-profile--view-favourites)
-    (define-key map (kbd "K") #'mastodon-profile--view-bookmarks)
+    (define-key map (kbd "O")      #'mastodon-profile--my-profile)
+    (define-key map (kbd "U")      #'mastodon-profile--update-user-profile-note)
+    (define-key map (kbd "V")      #'mastodon-profile--view-favourites)
+    (define-key map (kbd "K")      #'mastodon-profile--view-bookmarks)
+    (define-key map (kbd ":")      #'mastodon-user-settings)
     ;; minor views
-    (define-key map (kbd "R") #'mastodon-views--view-follow-requests)
-    (define-key map (kbd "S") #'mastodon-views--view-scheduled-toots)
-    (define-key map (kbd "I") #'mastodon-views--view-filters)
-    (define-key map (kbd "G") #'mastodon-views--view-follow-suggestions)
-    (define-key map (kbd "X") #'mastodon-views--view-lists)
-    (define-key map (kbd "SPC") #'mastodon-tl--scroll-up-command)
+    (define-key map (kbd "R")      #'mastodon-views--view-follow-requests)
+    (define-key map (kbd "S")      #'mastodon-views--view-scheduled-toots)
+    (define-key map (kbd "I")      #'mastodon-views--view-filters)
+    (define-key map (kbd "G")      #'mastodon-views--view-follow-suggestions)
+    (define-key map (kbd "X")      #'mastodon-views--view-lists)
+    (define-key map (kbd "SPC")    #'mastodon-tl--scroll-up-command)
+    (define-key map (kbd "!")      #'mastodon-tl--fold-post-toggle)
+    (define-key map (kbd "z")      #'bury-buffer)
     map)
   "Keymap for `mastodon-mode'.")
 
@@ -264,7 +319,9 @@ See `mastodon-toot-display-orig-in-reply-buffer'.")
 
 ;;;###autoload
 (defun mastodon ()
-  "Connect Mastodon client to `mastodon-instance-url' instance."
+  "Connect client to `mastodon-instance-url' instance.
+If there are any open mastodon.el buffers, switch to one instead.
+Prority in switching is given to timeline views."
   (interactive)
   (let* ((tls (list "home"
                     "local"
@@ -276,35 +333,55 @@ See `mastodon-toot-display-orig-in-reply-buffer'.")
                                 (get-buffer (concat "*mastodon-" el "*")))
                               tls) ; return first buff that exists
                      (cl-some (lambda (x)
-                                (when
-                                    (string-prefix-p "*mastodon-" (buffer-name x))
+                                (when (string-prefix-p "*mastodon-"
+                                                       (buffer-name x))
                                   (get-buffer x)))
                               (buffer-list))))) ; catch any other masto buffer
-    (mastodon-return-credential-account :force)
     (if buffer
-        (switch-to-buffer buffer)
+        (pop-to-buffer buffer '(display-buffer-same-window))
+      ;; we need to update credential-account in case setting have been changed
+      ;; outside mastodon.el in the meantime:
+      (mastodon-return-credential-account :force)
       (mastodon-tl--get-home-timeline)
-      (message "Loading Mastodon account %s on %s..."
+      (message "Loading fediverse account %s on %s..."
                (mastodon-auth--user-acct)
                mastodon-instance-url))))
 
 (defvar mastodon-profile-credential-account nil)
 
+;; TODO: the get request in mastodon-http--get-response often returns nil
+;; after waking from sleep, not sure how to fix, or if just my pc.
+;; interestingly it only happens with this function tho.
 (defun mastodon-return-credential-account (&optional force)
   "Return the CredentialAccount entity.
 Either from `mastodon-profile-credential-account' or from the
-server.
-FORCE means to fetch from the server and update
+server if that var is nil.
+FORCE means to fetch from the server in any case and update
 `mastodon-profile-credential-account'."
-  (let ((req '(mastodon-http--get-json
-               (mastodon-http--api "accounts/verify_credentials")
-               nil :silent)))
-    (if force
-        (setq mastodon-profile-credential-account
-              (eval req))
-      (or mastodon-profile-credential-account
-          (setq mastodon-profile-credential-account
-                (eval req))))))
+  (if (or force (not mastodon-profile-credential-account))
+      (setq mastodon-profile-credential-account
+            ;; TODO: we should signal a quit condition after 5 secs here
+            (condition-case nil
+                (mastodon-http--get-json
+                 (mastodon-http--api "accounts/verify_credentials")
+                 nil :silent)
+              (t ; req fails, return old value
+               mastodon-profile-credential-account)))
+    ;; else just return the var:
+    mastodon-profile-credential-account))
+
+(defvar mastodon-instance-data nil
+  "Instance data from the instance endpoint.")
+
+(defun mastodon-instance-data ()
+  "Return `mastodon-instnace-data' or else fetch from instance endpoint."
+  (or mastodon-instance-data
+      (setq mastodon-instance-data
+            (mastodon-http--get-json (mastodon-http--api "instance")))))
+
+(defun mastodon-instance-version ()
+  "Return the version string of user's instance."
+  (alist-get 'version (mastodon-instance-data)))
 
 ;;;###autoload
 (defun mastodon-toot (&optional user reply-to-id reply-json)
@@ -316,86 +393,102 @@ If REPLY-JSON is the json of the toot being replied to."
   (mastodon-toot--compose-buffer user reply-to-id reply-json))
 
 ;;;###autoload
-(defun mastodon-notifications-get (&optional type buffer-name force)
+(defun mastodon-notifications-get (&optional type buffer-name max-id)
   "Display NOTIFICATIONS in buffer.
 Optionally only print notifications of type TYPE, a string.
 BUFFER-NAME is added to \"*mastodon-\" to create the buffer name.
-FORCE means do not try to update an existing buffer, but fetch
-from the server and load anew."
+MAX-ID is a request parameter for pagination."
   (interactive)
-  (let ((buffer (if buffer-name
-                    (concat "*mastodon-" buffer-name "*")
-                  "*mastodon-notifications*")))
-    (if (and (not force)
-             (get-buffer buffer))
-        (progn (switch-to-buffer buffer)
-               (mastodon-tl--update))
-      (message "Loading your notifications...")
-      (mastodon-tl--init-sync (or buffer-name "notifications")
-                              "notifications"
-                              'mastodon-notifications--timeline
-                              type)
-      (with-current-buffer buffer
-        (use-local-map mastodon-notifications--map)))))
+  (let* ((buffer-name (or buffer-name "notifications"))
+         (buffer (concat "*mastodon-" buffer-name "*")))
+    (message "Loading your notifications...")
+    (mastodon-tl--init-sync
+     buffer-name
+     "notifications"
+     'mastodon-notifications--timeline
+     type
+     (when max-id
+       `(("max_id" . ,(mastodon-tl--buffer-property 'max-id))))
+     nil nil nil
+     (if (or (not mastodon-group-notifications)
+             ;; if version less than 1st grouped notifs release:
+             (> 4.3 (string-to-number
+                     (mastodon-instance-version))))
+         "v1"
+       "v2"))
+    (with-current-buffer (get-buffer-create buffer)
+      (use-local-map mastodon-notifications--map))
+    (message "Loading your notifications... Done")))
 
 ;; URL lookup: should be available even if `mastodon.el' not loaded:
 
 ;;;###autoload
-(defun mastodon-url-lookup (&optional query-url)
-  "If a URL resembles a mastodon link, try to load in `mastodon.el'.
-Does a WebFinger lookup.
+(defun mastodon-url-lookup (&optional query-url force)
+  "If a URL resembles a fediverse link, try to load in `mastodon.el'.
+Does a WebFinger lookup on the server.
 URL can be arg QUERY-URL, or URL at point, or provided by the user.
 If a status or account is found, load it in `mastodon.el', if
-not, just browse the URL in the normal fashion."
+not, just browse the URL in the normal fashion.
+If FORCE, do a lookup regardless of the result of `mastodon--fedi-url-p'."
   (interactive)
   (let* ((query (or query-url
-                    (thing-at-point-url-at-point)
                     (mastodon-tl--property 'shr-url :no-move)
+                    (thing-at-point-url-at-point)
                     (read-string "Lookup URL: "))))
-    (if (not (mastodon--fedi-url-p query))
+    (if (and (not force)
+             (not (mastodon--fedi-url-p query)))
         ;; (shr-browse-url query) ; doesn't work (keep our shr keymap)
-        (browse-url query)
+        (progn (message "Using external browser")
+               (browse-url query))
       (message "Performing lookup...")
       (let* ((url (format "%s/api/v2/search" mastodon-instance-url))
              (params `(("q" . ,query)
                        ("resolve" . "t"))) ; webfinger
              (response (mastodon-http--get-json url params :silent)))
-        (cond ((not (seq-empty-p
-                     (alist-get 'statuses response)))
+        (cond ((not (seq-empty-p (alist-get 'statuses response)))
                (let* ((statuses (assoc 'statuses response))
                       (status (seq-first (cdr statuses)))
                       (status-id (alist-get 'id status)))
-                 (mastodon-tl--thread status-id)))
-              ((not (seq-empty-p
-                     (alist-get 'accounts response)))
+                 (mastodon-tl--thread-do status-id)))
+              ((not (seq-empty-p (alist-get 'accounts response)))
                (let* ((accounts (assoc 'accounts response))
                       (account (seq-first (cdr accounts))))
                  (mastodon-profile--make-author-buffer account)))
               (t
+               (message "Lookup failed. Using external browser")
                (browse-url query)))))))
 
+(defun mastodon-url-lookup-force ()
+  "Call `mastodon-url-lookup' without checking if URL is fedi-like."
+  (interactive)
+  (mastodon-url-lookup nil :force))
+
 (defun mastodon--fedi-url-p (query)
   "Check if QUERY resembles a fediverse URL."
   ;; calqued off https://github.com/tuskyapp/Tusky/blob/c8fc2418b8f5458a817bba221d025b822225e130/app/src/main/java/com/keylesspalace/tusky/BottomSheetActivity.kt
   ;; thx to Conny Duck!
+  ;; mastodon at least seems to allow only [a-z0-9_] for usernames, plus "."
+  ;; but not at beginning or end, see https://github.com/mastodon/mastodon/issues/6830
+  ;; objects may have - in them
   (let* ((uri-parsed (url-generic-parse-url query))
          (query (url-filename uri-parsed)))
     (save-match-data
       (or (string-match "^/@[^/]+$" query)
           (string-match "^/@[^/]+/[[:digit:]]+$" query)
-          (string-match "^/user[s]?/@?[[:alnum:]]+$" query) ; @: pleroma or soapbox
+          (string-match "^/user[s]?/@?[[:alnum:]_]+$" query) ; @: pleroma or soapbox
           (string-match "^/notice/[[:alnum:]]+$" query)
           (string-match "^/objects/[-a-f0-9]+$" query)
           (string-match "^/notes/[a-z0-9]+$" query)
           (string-match "^/display/[-a-f0-9]+$" query)
-          (string-match "^/profile/[[:alpha:]]+$" query)
-          (string-match "^/p/[[:alpha:]]+/[[:digit:]]+$" query)
-          (string-match "^/[[:alpha:]]+$" query)
-          (string-match "^/u/[[:alpha:]]+$" query)
-          (string-match "^/c/[[:alnum:]]+$" query)
+          (string-match "^/profile/[[:alpha:]_]+$" query)
+          (string-match "^/p/[[:alpha:]_]+/[[:digit:]]+$" query)
+          (string-match "^/[[:alpha:]_]+$" query)
+          (string-match "^/u/[[:alpha:]_]+$" query)
+          (string-match "^/c/[[:alnum:]_]+$" query)
           (string-match "^/post/[[:digit:]]+$" query)
           (string-match "^/comment/[[:digit:]]+$" query) ; lemmy
-          (string-match "^/user[s]?/[[:alnum:]]+/statuses/[[:digit:]]+$" query) ; hometown
+          (string-match "^/@[^/]+/statuses/[[:alnum:]]" query) ; GTS
+          (string-match "^/user[s]?/[[:alnum:]_]+/statuses/[[:digit:]]+$" query) ; hometown
           (string-match "^/notes/[[:alnum:]]+$" query))))) ; misskey post
 
 (defun mastodon-live-buffers ()
@@ -420,27 +513,37 @@ Calls `mastodon-tl--get-buffer-type', which see."
 (defun mastodon-switch-to-buffer ()
   "Switch to a live mastodon buffer."
   (interactive)
-  (let* ((bufs (mastodon-live-buffers))
-         (buf-names (mapcar #'buffer-name bufs))
-         (choice (completing-read "Switch to mastodon buffer: "
-                                  buf-names)))
+  (let ((choice (completing-read
+                 "Switch to mastodon buffer: "
+                 (mapcar #'buffer-name (mastodon-live-buffers))
+                 nil :match)))
     (switch-to-buffer choice)))
 
+(defun mastodon--url-at-point ()
+  "`thing-at-point' provider function."
+  (get-text-property (point) 'shr-url))
+
 (defun mastodon-mode-hook-fun ()
   "Function to add to `mastodon-mode-hook'."
-  (when (require 'emojify nil :noerror)
+  (when (and mastodon-use-emojify
+             (require 'emojify nil :noerror))
     (emojify-mode t)
     (when mastodon-toot--enable-custom-instance-emoji
       (mastodon-toot--enable-custom-emoji)))
   (mastodon-profile--fetch-server-account-settings)
-  (when mastodon-tl--highlight-current-toot
-    (cursor-face-highlight-mode))) ; 29.1
+  (when (and mastodon-tl--highlight-current-toot
+             (fboundp #'cursor-face-highlight-mode))
+    (cursor-face-highlight-mode)) ; 29.1
+  ;; make `thing-at-point' functions work:
+  (setq-local thing-at-point-provider-alist
+              (append thing-at-point-provider-alist
+                      '((url . mastodon--url-at-point)))))
 
 ;;;###autoload
 (add-hook 'mastodon-mode-hook #'mastodon-mode-hook-fun)
 
 (define-derived-mode mastodon-mode special-mode "Mastodon"
-  "Major mode for Mastodon, the federated microblogging network."
+  "Major mode for fediverse services using the Mastodon API."
   (read-only-mode 1))
 
 (provide 'mastodon)
